%{
/*****************************************************************************
	TRIE dictionary translator version 3.0
						'89/07/22
						'92/07/27

	by TOKUNAGA Takenobu
	Tanaka Lab. Department of Computer Science
	Tokyo Institute of Technology
******************************************************************************/
#include <stdio.h>
#include <ctype.h>

#define GETCH()		ch = getc(fi)
#define PUTCH(p, char)	*p++ = char
#define COPYCH(p, q)	*p++ = *q++

#define TRUE	1
#define FALSE	0
#define ESC	27
#define	MAX_LINE_LEN	1000		/* max length of one rule	*/
#define	MAX_SYM_NO	500		/* max number of symbol per rule*/
#define HEAD_CH		"\001"
#define TAIL_CH		"\377"


FILE	*fi;				/* input file pointer		*/
FILE	*fo;				/* output file pointer		*/
char	*myname;			/* save argv[0] to this area	*/

int	ch;				/* look ahead character		*/
unsigned char	buffer[MAX_LINE_LEN];	/* buffer for a rule		*/
unsigned char	yytext[MAX_LINE_LEN];	/* buffer for 'yylval'		*/
unsigned char	*bp = buffer;		/* pointer to 'buffer'		*/
unsigned char	*yp = yytext;		/* pointer to 'yytext'		*/

unsigned char	*symbol[MAX_SYM_NO];	/* symbols in rules stored	*/
int	sym_n = 0, sym_i = 0;		/* pointer to symbol		*/

int	error_no= 0;			/* error counter		*/

struct argument {
    unsigned char	*info;		/* argument information	    	*/
    struct argument	*succ;		/* pointer to successor	    	*/
};

struct word_cell {
    unsigned char	*sym;		/* symbol name			*/
    struct argument	*arg;		/* arguments			*/
    struct word_cell	*succ;		/* pointer to successor		*/
    struct word_cell	*next;		/* pointer to next entry	*/
}	tail = {(unsigned char *)TAIL_CH,NULL,NULL,NULL},
	head = {(unsigned char *)HEAD_CH,NULL,NULL,&tail};
					/* sentinels			*/
struct word_cell    *wp,*root = &head;	/* root and work pointers	*/

char	*malloc();
void	go_succ();
void	go_next();
void	yyerror();

%}
%union	{ unsigned char	*string; }
%start dictionary
%token <string>	ARGUMENTS ARROW KANJI ASCII NUMBER PROGRAM SYMBOL
%type  <string> cItem cList headSpec ident idChar idChars '_'
%% 
dictionary:
	|   dictionary rule
;
rule	:   headCat ARROW body '.'   {
		wp = root;
		sym_i = 1;
		go_next();
		sym_n = 0;
	    }
	|   error   {
		sync();
		yyerrok;
		yyclearin;	
	    }
;
body	:   term
	|   body ',' term
;
term	:   '[' words ']'
	|   PROGRAM	{
		symbol[sym_n++] = StrPrt("program([%s])",$1);
	    }
	|   cat
;
word	:   SYMBOL	{
		symbol[sym_n++] = $1;
	    }
	|   KANJI	{
		break_kanji($1);
	    }
	|   ASCII	{
		symbol[sym_n++] = $1;
	    }
	|   NUMBER	{
		symbol[sym_n++] = $1;
	    }
;
words	:   word
	|   words  word
	|   words ',' word
;
headSpec:   '<' cList '>'	{
		$$ = StrPrt("[%s]",$2);
		free($2);
	    }
;
cList	:   cItem
	|   cList ',' cItem	{
		$$ = StrPrt("%s,%s",$1,$3);
		free($1);
		free($3);
	    }
;
cItem	:   SYMBOL	/* exclude morph type written in KANJI */
	|   ASCII
	|   '*' ident		{
	    	$$ = StrPrt("head(%s)",$2);
		free($2);
	    }
	|   '_'		{
		$$ = (unsigned char *)"ignore(me)";
	    }
;
headCat	:   ident   {
		symbol[sym_n++] = StrPrt("[%s,[],[head(%s)]]",$1,$1);
	    }
	|   ident headSpec	{
		symbol[sym_n++] = StrPrt("[%s,[],%s]",$1,$2);
	    }
	|   ident ARGUMENTS	{
		symbol[sym_n++] = StrPrt("[%s,[%s],[head(%s)]]",$1,$2,$1);
	    }
	|   ident ARGUMENTS headSpec	{
		symbol[sym_n++] = StrPrt("[%s,[%s],%s]",$1,$2,$3);
	    }
;
cat	:   ident   {
		symbol[sym_n++] = StrPrt("[%s]",$1);
	    }
	|   ident ARGUMENTS	{
		symbol[sym_n++] = StrPrt("[%s,%s]",$1,$2);
	    }
;
ident	:   SYMBOL
	|   ASCII idChars {
	    	$$ = StrPrt("%s%s",$1,$2);
	    }
	|   KANJI idChars {
	    	$$ = StrPrt("%s%s",$1,$2);
	    }
;
idChars	:   {	$$ = (unsigned char *)""; }
	|   idChars idChar {
	  	$$ = StrPrt("%s%s",$1,$2);
	    }
;
idChar	:   KANJI
	|   ASCII
	|   NUMBER
  	|   '_'
;
%%

/*
  read_a_rule():
  reads upto period from the input file to the buffer.
  convert 7bit JIS kanji code to 8bit EUC if any.
*/
int
read_a_rule()
{
    static int state = 0;   /* the state of the definite finite automaton */

    bp = buffer;	    /* reset the buffer pointer */
    for (;;) {
	switch (state) {
	case 0 :
	    switch (ch) {
	    case EOF  : return(EOF);
	    case '/'  : state = 1;
		    	break;
	    case '%'  : state = 4;
		    	break;
	    case '\'' : state = 5;
		    	PUTCH(bp,ch);
		    	break;
	    case ESC  : state = 6;
		    	break;
	    case '.'  : state = 11;
		    	PUTCH(bp,ch);
		    	break;
	    default   : if (ch >= ' ') PUTCH(bp,ch);
	    }
	    GETCH(); break;
	case 1 :
	    switch (ch) {
	    case '*'  : state = 2;
		    	GETCH();
		    	break;
	    default   : state = 0;
		    	PUTCH(bp,'/');
	    }
	    break;
	case 2 :
	    switch (ch) {
	    case '*'  : state = 3;
			break;
	    }
	    GETCH(); break;
	case 3 :
	    switch (ch) {
	    case '*'  : state = 3;
		    	break;
	    case '/'  : state = 0;
		    	break;
	    default   : state = 2;
	    }
	    GETCH(); break;
	case 4 :
	    switch (ch) {
	    case '\n' : state = 0;
		    	break;
	    default   : state = 4;
	    }
	    GETCH(); break;
	case 5 :
	    switch (ch) {
	    case '\'' : state = 0;
		    	PUTCH(bp,ch);
		    	break;
	    case '\n' : *bp='\0';
		    	yyerror("missing closing quote?");
		    	return(0);
	    default   : PUTCH(bp,ch);
	    }
	    GETCH(); break;
	case 6 :
	    switch (ch) {
	    case '$'  : state = 7;
		    	break;
	    default   : *bp='\0';
		    	yyerror("irregal escape sequence?");
		    	return(0);
	    }
	    GETCH(); break;
	case 7:
	    switch (ch) {
	    case '@'  :
	    case 'B'  : state = 8;
		    	break;
	    default   : *bp='\0';
		    	yyerror("irregal escape sequence?");
		    	return(0);
	    }
	    GETCH(); break;
	case 8 :
	    switch (ch) {
	    case ESC  : state = 9;
		    	break;
	    default   : state = 8;
		    	PUTCH(bp,ch|0x80);	/* convert to EUC */
	    }
	    GETCH(); break;
	case 9 :
	    switch (ch) {
	    case '('  :	state = 10;
		    	break;
	    default   : *bp='\0';
		    	yyerror("irregal escape sequence?");
		    	return(0);
	    }
	    GETCH(); break;
	case 10 :
	    switch (ch) {
	    case 'B'  :
	    case 'H'  :
	    case 'J'  : state = 0;
		    	break;
	    default   : *bp='\0';
		    	yyerror("irregal escape sequence?");
		    	return(0);
	    }
	    GETCH(); break;
	case 11 :
	    state = 0;
	    if (ch<=' ') {
		PUTCH(bp,'\0');
		bp = buffer;    /* reset pointer to 'buffer' */
		return(0);
	    }
	}
    }
}

void
copy_between_quotes()
{
    while (*bp=='\'') {
	for (COPYCH(yp,bp); ((*bp!='\'')&&(*bp!='\0')); COPYCH(yp,bp));
	COPYCH(yp,bp);
    }
}

void
copy_upto(lch,rch)		/* copy 'buffer' to 'yylval' until "rch"*/
    unsigned char lch,rch;	/* appears considering nesting of "lch" */
				/* and "rch" pairing */
{
    while ((*bp!=rch)&&(*bp!='\0')) {
	if (*bp==lch) {
	    COPYCH(yp,bp); copy_upto(lch, rch);
	}
	else if (*bp=='\'') {
	    copy_between_quotes();
	}
	else {
	    COPYCH(yp,bp);
	}
    }
    COPYCH(yp,bp);
}

unsigned char *
StrAlc(string)
    unsigned char    *string;
{
    unsigned char    *p;

    if ((p = (unsigned char *)malloc(strlen(string)+1))==NULL) {
	fprintf(stderr,"%s: memory shotage in StrAlc.\n",myname);
	exit(1);
    }
    else {
	strcpy(p,string);
	return(p);
    }
}
    
unsigned char *
StrPrt(fmt,a,b,c,d,e,f,g,h)
    char    *fmt;
    int	    a,b,c,d,e,f,g,h;
{
    unsigned char    tmp[MAX_LINE_LEN];
    unsigned char    *p;

    sprintf(tmp,fmt,a,b,c,d,e,f,g,h);
    if ((p = (unsigned char *)malloc(strlen(tmp)+1))==NULL) {
	fprintf(stderr,"%s: memory shotage in StrPrt.\n",myname);
	exit(1);
    }
    else {
	strcpy(p,tmp);
	return(p);
    }
}

/*
StrCmp(s1, s2): returns
	-1 if s1 < s2
	 0 if s1 = s2
	 1 if s1 < s2.
*/
	  
int
StrCmp(s1, s2)
    unsigned char *s1, *s2;
{
    while (*s1!='\0' && *s2!='\0') {
	if (*s1 < *s2) return(-1);
	else if (*s1 > *s2) return (1);
	else {
	    s1++;
	    s2++;
	}
    }
    if (*s1==0 && *s2==0) return(0);
    else if (*s1==0) return(-1);
    else return(1);
}

/*
  IsEuc(ch): test if 'ch' is EUC code.
*/
int
IsEuc(ch)
    unsigned char ch;
{
    return(ch&0x80);
}

/*
  IsBlank(ch): test if 'ch' is space or tab.
*/
int
IsBlank(ch)
    unsigned char ch;
{
    return((ch==0x20||ch==0x09));
}

int
yylex()		/* lexical analyzer called by yyparse */
{
    int  token_type;
    
    if (*bp=='\0')
	if (read_a_rule()==EOF) return(EOF);
    yp = yytext;    /* reset pointer to 'yylval.string' */

    while (IsBlank(*bp)) bp++;
    if (IsEuc(*bp)) {
	for (COPYCH(yp,bp); IsEuc(*bp); COPYCH(yp,bp));
	token_type = KANJI;
    }
    else if (islower(*bp)) {
	for (COPYCH(yp,bp); (isalnum(*bp)||(*bp=='_'));
	     COPYCH(yp,bp));
	token_type = ASCII;
    }
    else if (isdigit(*bp)) {
	for (COPYCH(yp,bp); isdigit(*bp); COPYCH(yp,bp));
	token_type = NUMBER;
    }
    else if (*bp=='\'') {
	copy_between_quotes();
	token_type = SYMBOL;
    }
    else if (*bp == '{') {
	bp++;				/* skip "{" */
	copy_upto('{', '}');
	yp--;				/* delete "}' */
	token_type = PROGRAM;
    }
    else if (*bp == '(') {
	bp++;				/* skip "(" */
	copy_upto('(', ')');
	yp--;				/* delete ")" */
	token_type = ARGUMENTS;
    }
    else if ((*bp=='-')&&(*(bp+1)=='-')&&(*(bp+2)=='>')) {
	COPYCH(yp,bp); COPYCH(yp,bp); COPYCH(yp,bp);
	token_type = ARROW;
    }
    else if ((*bp=='.')&&(*(bp+1)=='\0')) {
	token_type = *bp;
	COPYCH(yp,bp);
    }
    else {
	token_type = *bp;
	COPYCH(yp,bp);
    }
    *yp = '\0';
    yylval.string = StrAlc(yytext);
    return(token_type);
}

void
break_kanji(p)
    char *p;	/* points the first byte of a KANJI sequence */
{
    while (*p!='\0') {
	symbol[sym_n++] = StrPrt("%c%c\0",*p,*(p+1));
	p++;
	p++;
    }
	
}

void
go_next()
{
    struct word_cell *p;

    while(StrCmp(symbol[sym_i],wp->next->sym)>=0) wp = wp->next;
    if (StrCmp(symbol[sym_i],wp->sym)) {
	p = (struct word_cell *)malloc(sizeof(struct word_cell));
	p->sym   = symbol[sym_i];
	p->arg   = NULL;
	p->succ  = NULL;
	p->next  = wp->next;
	wp->next = p;
	wp	 = p;
    }
    go_succ();
}

void
go_succ()
{
    struct word_cell *p;
    struct argument  *q;

    if (sym_i==(sym_n-1)) {
	q = (struct argument *)malloc(sizeof(struct argument));
	q->succ = wp->arg;
	q->info = symbol[0];
	wp->arg = q;
    }
    else {
	if (wp->succ==NULL) {
	    p = (struct word_cell *)malloc(sizeof(struct word_cell));
	    p->sym   = (unsigned char *)HEAD_CH;
	    p->arg   = NULL;
	    p->succ  = NULL;
	    p->next  = &tail;
	    wp->succ = p;
	}
	sym_i++;
	wp = wp->succ;
	go_next();
    }
}

void
tabs(n)
    int n;
{
    for (;n > 0; n--) fprintf(fo,"   ");
}

void
write_dict(p,nest)
    int		nest;
    struct word_cell *p;
{
    struct argument  *q;

    if (p!=NULL) {
	tabs(nest);
	if (nest==0)	fprintf(fo,"trie(%s,",p->sym);
	else		fprintf(fo,"[%s,",p->sym);

	/* print argument list */
	fprintf(fo,"[");    /* left bracket of argument list */
	if ((p->arg)!=NULL) {
	    q = p->arg;
	    fprintf(fo,"%s",q->info);
	    for (q=q->succ;q!=NULL;q=q->succ) fprintf(fo,",%s",q->info);
	}
	fprintf(fo,"],");   /* right bracket of argument list */

	/* print trailing symbols */
	if (p->succ==NULL) {
	    fprintf(fo,"[]");
	}
	else {
	    fprintf(fo,"[\n");
	    write_dict(p->succ->next,nest+1);
	    fprintf(fo,"]");
	}

	/* print closing bracket */
	if (nest==0) {
	    fprintf(fo,").\n");
	}
	else {
	    if (p->next==(&tail))   fprintf(fo,"]");
	    else		    fprintf(fo,"],\n");
	}

	/* write next entry */
	if (p->next!=(&tail)) write_dict(p->next,nest);
    }
}

void
yyerror(s)		/* print error message and error location */
    char *s;
{
    unsigned char *p;

    error_no++;
    fprintf(stderr,"%s\n",s);
    fprintf(stderr,">>> %s\n",buffer);
    fprintf(stderr,">>> ");
    for (p=buffer; p<(bp-1); p++) fprintf(stderr,".");
    fprintf(stderr,"^\n\n");
}

void
sync()		/* easy syncronizing routine : skip current rule */
{		/* see head of "read_a_rule" why this works well */
    *bp='\0';
}

main(argc,argv)
    int argc;
    char *argv[];
{

    int c;

    myname = argv[0];

    if (argc!=3) {
	fprintf(stderr,"%s: exactly 2 arguments required\n",myname);
	exit(1);
    }

    /* open input files */
    if (strcmp(argv[1],"STDIN")) {
	if ((fi = fopen(argv[1],"r"))==NULL) {
	    fprintf(stderr,"%s: can't open %s\n",myname,argv[1]);
	    exit(1);
	}
    }
    else {
	fi = stdin;
    }
    /* output files */
    if (strcmp(argv[2],"STDOUT")) {
	if ((fo = fopen(argv[2],"w"))==NULL) {
	    fprintf(stderr,"%s: can't open %s\n",myname,argv[2]);
	    exit(1);
	}
    }
    else {
	fo = stdout;
    }

    /* start translation */
    *bp = '\0';
    GETCH();
    if (yyparse()) {
	fprintf(stderr,"%s: fatal error occured. can't recover.\n",myname);
	exit(1);
    }
    /* write TRIE dictionary to file */
    write_dict(root->next,0);

    fclose(fo);
    fclose(fi);

    /* diagnostics */
    switch (error_no) {
    case 0 :	fprintf(stderr,"No error.\n"); break;
    case 1 :	fprintf(stderr," 1 error.\n"); break;
    default :   fprintf(stderr,"%d errors.\n",error_no);
    }
    exit(0);
}
